/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 by the Free Software Foundation.               *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
 \* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <stdio.h>
#include "incremental_board.h"


#if 0
| This file contains code for incremental board updates that keeps track
| of the following information for each string:
| 
| * The color of the string.
| 
| * Number of stones in the string.
| 
| * Origin of the string, i.e. a canonical reference point, defined
| to be the stone with smallest i coordinate and if there is a tie with
| smallest j coordinate.
| 
| * A list of the stones in the string.
| 
| * Number of liberties.
| 
| * A list of the liberties. If there are too many liberties the list is
| truncated.
| 
| * The number of neighbor strings.
| 
| * A list of the neighbor strings.
| 
| 
| The basic data structure is
| 
| struct string_data {
|   int color;                       /* Color of string, BLACK or WHITE */
|   int size;                        /* Number of stones in string. */
|   int origini;                     /* Coordinates of "origin", i.e. */
|   int originj;                     /* "upper left" stone. */
|   int liberties;                   /* Number of liberties. */
|   int libi[MAX_LIBERTIES];         /* Coordinates of liberties. */
|   int libj[MAX_LIBERTIES];
|   int neighbors;                   /* Number of neighbor strings */
|   int neighborlist[MAXCHAIN];      /* List of neighbor string numbers. */
|   int mark;                        /* General purpose mark. */
| };
| 
| struct string_data string[MAX_STRINGS];
| 
| It should be clear that almost all information is stored in the
| 'string' array. To get a mapping from the board coordinates to the
| 'string' array we have
| 
| int string_number[MAX_BOARD][MAX_BOARD];
| 
| which contains indices into the 'string' array. This information is only
| valid at nonempty vertices, however, so it is necessary to first
| verify that p[i][j] != EMPTY.
| 
| The 'string_data' structure does not include an array of the stone
| coordinates. This information is stored in a separate array (or rather
| two):
| 
| int next_stonei[MAX_BOARD][MAX_BOARD];
| int next_stonej[MAX_BOARD][MAX_BOARD];
| 
| These arrays implement cyclic linked lists of stones. Each vertex
| contains a pointer to another (possibly the same) vertex. Starting at
| an arbitrary stone on the board, following these pointers should
| traverse the entire string in an arbitrary order before coming back to
| the starting point. As for the 'string_number' array, this information
| is invalid at empty points on the board. This data structure has the
| good properties of requiring fixed space (regardless of the number of
| strings) and making it easy to add a new stone or join two strings.
| 
| Additionally the code makes use of some some work variables:
| 
| static int ml[MAX_BOARD][MAX_BOARD];
| static int liberty_mark;
| static int string_mark;
| static int next_string;
| static int strings_initialized = 0;
| 
| The 'ml' array and 'liberty_mark' are used to "mark" liberties on the
| board, e.g. to avoid counting the same liberty twice. The convention
| is that if ml[i][j] has the same value as 'liberty_mark', then (i, j)
| is marked. To clear all marks it suffices to increase the value of
| 'liberty_mark', since it is never allowed to decrease.
| 
| The same relation holds between the 'mark' field of the 'string_data'
| structure and 'string_mark'. Of course these are used for marking
| individual strings.
| 
| 'next_string' gives the number of the next available entry in the
| 'string' array. 'strings_initialized' is set to one when all data
| structures are known to be up to date. Given an arbitrary board
| position in the 'p' array, this is done by calling
| 'incremental_board_init()'. It is not necessary to call this function
| explicitly since any other function that needs the information does
| this if it has not been done.
| 
| The interesting part of the code is the incremental update of the data
| structures when a stone is played and subsequently removed. To
| understand the strategies involved in adding a stone it is necessary
| to first know how undoing a move works. The idea is that as soon as
| some piece of information is about to be changed, the old value is
| pushed onto a stack which stores the value and its address. The stack
| is built from the following structures:
| 
| struct change_stack_entry {
|   int *address;
|   int value;
| };
| 
| struct change_stack_entry change_stack[STACK_SIZE];
| int change_stack_index;
| 
| and manipulated with the macros
| 
| BEGIN_CHANGE_RECORD()
| PUSH_VALUE(v)
| POP_MOVE()
| 
| Calling 'BEGIN_CHANGE_RECORD()' stores a null pointer in the address
| field to indicate the start of changes for a new move. As mentioned
| earlier 'PUSH_VALUE()' stores a value and its corresponding address.
| Assuming that all changed information has been duly pushed onto the
| stack, undoing the move is only a matter of calling 'POP_MOVE()',
| which simply assigns the values to the addresses in the reverse order
| until the null pointer is reached. This description is slightly
| simplified because this stack can only store 'int' values and we need
| to also store changes to the board. Thus we have two parallel stacks
| where one stores 'int' values and the other one stores 'board_t'
| values.
| 
| When a new stone is played on the board, first captured opponent
| strings, if any, are removed. In this step we have to push the board
| values and the 'next_stone' pointers for the removed stones, and
| update the liberties and neighbor lists for the neighbors of the
| removed strings. We do not have to push all information in the
| 'string' entries of the removed strings however. As we do not reuse
| the entries they will remain intact until the move is pushed and they
| are back in use.
| 
| After this we put down the new stone and get three distinct cases:
| 
| 1. The new stone is isolated, i.e. it has no friendly neighbor.
| 2. The new stone has exactly one friendly neighbor.
| 3. The new stone has at least two friendly neighbors.
| 
| The first case is easiest. Then we create a new string by using the
| number given by 'next_string' and increasing this variable. The string
| will have size one, 'next_stone' points directly back on itself, the
| liberties can be found by looking for empty points in the four
| directions, possible neighbor strings are found in the same way, and
| those need also to remove one liberty and add one neighbor.
| 
| In the second case we do not create a new string but extend the
| neighbor with the new stone. This involves linking the new stone into
| the cyclic chain, if needed moving the origin, and updating liberties
| and neighbors. Liberty and neighbor information also needs updating
| for the neighbors of the new stone.
| 
| In the third case finally, we need to join already existing strings.
| In order not to have to store excessive amounts of information, we
| create a new string for the new stone and let it assimilate the
| neighbor strings. Thus all information about those can simply be left
| around in the 'string' array, exactly as for removed strings. Here it
| becomes a little more complex to keep track of liberties and neighbors
| since those may have been shared by more than one of the joined
| strings. Making good use of marks it all becomes rather
| straightforward anyway.
| 
| 
| Before delving into the code, it may be worth noticing that the often
| used construction
| 
|     FIRST_STONE(s, i, j);
|     do {
| 	...
| 	NEXT_STONE(i, j);
|     } while (!BACK_TO_FIRST_STONE(s, i, j));
| 
| traverses the stones of the string with number 's' exactly once, with
| (i, j) holding the coordinates. In general (i, j) are used as board
| coordinates and 's' as an index into the 'string' array or sometimes a
| pointer to an entry in the 'string' array.
#endif

/* Main array of string information. */
static struct string_data string[MAX_STRINGS];

/* Stacks and stack pointers. */
struct change_stack_entry change_stack[STACK_SIZE];
struct vertex_stack_entry vertex_stack[STACK_SIZE];
int change_stack_index;
int vertex_stack_index;

/* Index into list of strings. The index is only valid if there is a
 * stone at the vertex.
 */
static int string_number[MAX_BOARD][MAX_BOARD];

/* The stones in a string are linked together in a cyclic list. These
 * are the coordinates to the next stone in the string.
 */
static int next_stonei[MAX_BOARD][MAX_BOARD];
static int next_stonej[MAX_BOARD][MAX_BOARD];

/* Macros to traverse the stones of a string. */
#define FIRST_STONE(s, i, j) i = string[s].origini; j = string[s].originj
#define NEXT_STONE(i, j) do {\
    int nexti = next_stonei[i][j];\
    j = next_stonej[i][j];\
    i = nexti;\
  } while (0)
#define BACK_TO_FIRST_STONE(s, i, j) (i == string[s].origini &&\
				      j == string[s].originj)

/* Assorted useful macros. Some of them could have been functions but
 * are implemented as macros for speed.
 */
#define LIBERTY(i, j) (p[i][j] == EMPTY)
#define UNMARKED_LIBERTY(i, j) (p[i][j] == EMPTY && ml[i][j] != liberty_mark)
#define MARK_LIBERTY(i, j) ml[i][j] = liberty_mark

#define UNMARKED_STRING(i, j) (string[string_number[i][j]].mark != string_mark)
#define UNMARKED_OPPONENT_STRING(s, i, j)\
  (p[i][j] == OTHER_COLOR(string[s].color)\
   && string[string_number[i][j]].mark != string_mark)
#define UNMARKED_OWN_STRING(s, i, j)\
  (p[i][j] == string[s].color\
   && string[string_number[i][j]].mark != string_mark)
#define MARK_STRING(i, j) string[string_number[i][j]].mark = string_mark

#define LIBERTIES(i, j) string[string_number[i][j]].liberties
#define ADD_LIBERTY(s, i, j) do {\
    if (string[s].liberties < MAX_LIBERTIES) {\
      string[s].libi[string[s].liberties] = i;\
      string[s].libj[string[s].liberties] = j;\
    }\
    string[s].liberties++;\
  } while (0)
#define ADD_AND_MARK_LIBERTY(s, i, j) do {\
    if (string[s].liberties < MAX_LIBERTIES) {\
      string[s].libi[string[s].liberties] = i;\
      string[s].libj[string[s].liberties] = j;\
    }\
    string[s].liberties++;\
    ml[i][j] = liberty_mark;\
  } while (0)
#define ADD_NEIGHBOR(s, i, j)\
  string[s].neighborlist[string[s].neighbors++] = string_number[i][j]

#define DO_ADD_STONE(i, j, color) do {\
    PUSH_VERTEX(p[i][j]);\
    p[i][j] = color;\
    hashdata_invert_stone(&hashdata, i, j, color);\
  } while (0)
#define DO_REMOVE_STONE(i, j) do {\
    PUSH_VERTEX(p[i][j]);\
    hashdata_invert_stone(&hashdata, i, j, p[i][j]);\
    p[i][j] = EMPTY;\
  } while (0)


/* Are the data structures up to date for the current board position. */
static int strings_initialized = 0;

/* Number of the next free string. */
static int next_string;

/* For marking purposes. */
static int ml[MAX_BOARD][MAX_BOARD];
static int liberty_mark;
static int string_mark;


#if 0
/* Debug function. Dump all string information. */
static void
dump_incremental_board(void)
{
  int i, j, s;
  
  for (i=0; i<board_size; i++) {
    for (j=0; j<board_size; j++)
      if (p[i][j] == EMPTY)
	fprintf(stderr, " . ");
      else
	fprintf(stderr, "%2d ", string_number[i][j]);
    fprintf(stderr, "\n");
  }

  for (s=0; s<next_string; s++) {
    if (p[string[s].origini][string[s].originj] == EMPTY)
      continue;
    
    gprintf("%o%d %s %m size %d, %d liberties, %d neighbors\n", s,
	    string[s].color == WHITE ? "white" : "black",
	    string[s].origini, string[s].originj, string[s].size,
	    string[s].liberties, string[s].neighbors);
    gprintf("%ostones:");

    FIRST_STONE(s, i, j);
    do {
      gprintf("%o %m", i, j);
      NEXT_STONE(i, j);
    } while (!BACK_TO_FIRST_STONE(s, i, j));
    
    gprintf("%o\nliberties:");
    for (i=0; i<string[s].liberties; i++)
      gprintf("%o %m", string[s].libi[i], string[s].libj[i]);
    
    gprintf("%o\nneighbors:");
    for (i=0; i<string[s].neighbors; i++)
      gprintf("%o %d(%m)", string[s].neighborlist[i],
	      string[string[s].neighborlist[i]].origini,
	      string[string[s].neighborlist[i]].originj);
    gprintf("%o\n\n");
  }
}
#endif


/* Build a string and its cyclic list representation from scratch.
 * propagate_string(i, j, m, n) adds the stone (i, j) to the string
 * (m, n) and recursively continues with not already included friendly
 * neighbors. To start a new string at (i, j), use
 * propagate_string(i, j, i, j). The size of the string is returned.
 */

static int
propagate_string(int i, int j, int m, int n)
{
  int size = 1;
  int k;
  
  if (i == m && j == n) {
    /* Start a new string. */
    next_stonei[i][j] = i;
    next_stonej[i][j] = j;
  }
  else {
    /* Link the stone at (i, j) to the string including (m, n) */
    string_number[i][j] = string_number[m][n];
    next_stonei[i][j] = next_stonei[m][n];
    next_stonej[i][j] = next_stonej[m][n];
    next_stonei[m][n] = i;
    next_stonej[m][n] = j;
  }

  /* Look in all four directions for more stones to add. */
  for (k=0; k<4; k++) {
    int di = deltai[k];
    int dj = deltaj[k];
    if (ON_BOARD(i+di, j+dj)
	&& p[i+di][j+dj] == p[i][j]
	&& string_number[i+di][j+dj] == -1)
      size += propagate_string(i+di, j+dj, m, n);
  }
  
  return size;
}


/* Build the lists of liberties and neighbors of a string from
 * scratch. No information is pushed onto the stack by this function.
 */

static void
find_liberties_and_neighbors(int s)
{
  int i, j;

  /* Clear the marks. */
  liberty_mark++;
  string_mark++;

  /* Traverse the stones of the string, by following the cyclic chain. */
  FIRST_STONE(s, i, j);
  do {
    /* Look in each direction for new liberties or new neighbors. Mark
     * already visited liberties and neighbors.
     */
    if (i > 0) {
      if (UNMARKED_LIBERTY(i-1, j)) {
	ADD_AND_MARK_LIBERTY(s, i-1, j);
      }
      else if (UNMARKED_OPPONENT_STRING(s, i-1, j)) {
	ADD_NEIGHBOR(s, i-1, j);
	MARK_STRING(i-1, j);
      }
    }
    
    if (i < board_size-1) {
      if (UNMARKED_LIBERTY(i+1, j)) {
	ADD_AND_MARK_LIBERTY(s, i+1, j);
      }
      else if (UNMARKED_OPPONENT_STRING(s, i+1, j)) {
	ADD_NEIGHBOR(s, i+1, j);
	MARK_STRING(i+1, j);
      }
    }
    
    if (j > 0) {
      if (UNMARKED_LIBERTY(i, j-1)) {
	ADD_AND_MARK_LIBERTY(s, i, j-1);
      }
      else if (UNMARKED_OPPONENT_STRING(s, i, j-1)) {
	ADD_NEIGHBOR(s, i, j-1);
	MARK_STRING(i, j-1);
      }
    }
    
    if (j < board_size-1) {
      if (UNMARKED_LIBERTY(i, j+1)) {
	ADD_AND_MARK_LIBERTY(s, i, j+1);
      }
      else if (UNMARKED_OPPONENT_STRING(s, i, j+1)) {
	ADD_NEIGHBOR(s, i, j+1);
	MARK_STRING(i, j+1);
      }
    }

    NEXT_STONE(i, j);
  } while (!BACK_TO_FIRST_STONE(s, i, j));
}


/* Update the liberties of a string from scratch, first pushing the
 * old information.
 */

static void
update_liberties(int s)
{
  int i, j, k;

  /* Push the old information. */
  PUSH_VALUE(string[s].liberties);
  for (k=0; k<string[s].liberties; k++) {
    PUSH_VALUE(string[s].libi[k]);
    PUSH_VALUE(string[s].libj[k]);
  }
  string[s].liberties = 0;

  /* Clear the liberty mark. */
  liberty_mark++;

  /* Traverse the stones of the string, by following the cyclic chain. */
  FIRST_STONE(s, i, j);
  do {
    /* Look in each direction for new liberties. Mark already visited
     * liberties. 
     */
    if (i > 0 && UNMARKED_LIBERTY(i-1, j)) {
      ADD_AND_MARK_LIBERTY(s, i-1, j);
    }
    
    if (i < board_size-1 && UNMARKED_LIBERTY(i+1, j)) {
      ADD_AND_MARK_LIBERTY(s, i+1, j);
    }
    
    if (j > 0 && UNMARKED_LIBERTY(i, j-1)) {
      ADD_AND_MARK_LIBERTY(s, i, j-1);
    }
    
    if (j < board_size-1 && UNMARKED_LIBERTY(i, j+1)) {
      ADD_AND_MARK_LIBERTY(s, i, j+1);
    }

    NEXT_STONE(i, j);
  } while (!BACK_TO_FIRST_STONE(s, i, j));
}


/* Set up incremental board structures and populate them with the
 * strings available in the position given by p[][]. Clear the stacks
 * and start the mark numbers from zero. All undo information is lost
 * by calling this function.
 */

void
incremental_board_init()
{
  int i, j, s;
  next_string = 0;
  liberty_mark = 0;
  string_mark = 0;
  CLEAR_STACKS();

  memset(string, 0, sizeof(string));
  memset(ml, 0, sizeof(ml));

  /* propagate_string relies on non-assigned stones to have
   * string_number -1.
   */
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++)
      string_number[i][j] = -1;

  /* Find the existing strings. */
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++)
      if (p[i][j] != EMPTY && string_number[i][j] == -1) {
	string_number[i][j] = next_string;
	string[next_string].size = propagate_string(i, j, i, j);
	string[next_string].color = p[i][j];
	string[next_string].origini = i;
	string[next_string].originj = j;
	string[next_string].mark = 0;
	next_string++;
      }
  
  /* Fill in liberty and neighbor info. */
  for (s=0; s<next_string; s++) {
    find_liberties_and_neighbors(s);
  }

  /* Now we can trust the information. */
  strings_initialized = 1;
}


/* Don't trust the incremental string data until it's reinitialized.
 *
 * This function should be called if the board is modified by other
 * means than incremental_play_move() or incremental_undo_move().
 * It's also useful to force a recomputation of the strings if we
 * don't have any immediate plans to undo the move, because it recovers
 * undo stack space and holes in the 'string' array.
 */

void
incremental_invalidate_strings()
{
  strings_initialized = 0;
}


/* Remove a string from the list of neighbors and push the changed
 * information.
 */

static void
remove_neighbor(int str, int n)
{
  int k;
  int done=0;
  struct string_data *s = &string[str];
  for (k=0; k<s->neighbors; k++)
    if (s->neighborlist[k] == n) {
      /* We need to push the last entry too because it may become
       * destroyed later.
       */
      PUSH_VALUE(s->neighborlist[s->neighbors - 1]);
      PUSH_VALUE(s->neighborlist[k]);
      PUSH_VALUE(s->neighbors);
      s->neighborlist[k] = s->neighborlist[s->neighbors - 1];
      s->neighbors--;
      done = 1;
      break;
    }
  assert(done);
}


/* Add a new neighbor at the end of the list and push the changed
 * information.
 *
 * FIXME: This doesn't have to be conditional anymore.
 */

static void
conditional_add_neighbor(int str, int n)
{
  int k;
  struct string_data *s = &string[str];
  for (k=0; k<s->neighbors; k++)
    if (s->neighborlist[k] == n)
      abort(); /* Shouldn't happen. */
  PUSH_VALUE(s->neighbors);
  s->neighborlist[s->neighbors++] = n;
}


/* Remove one liberty from the list of liberties, pushing changed
 * information. If the string had more liberties than the size of the
 * list, rebuild the list from scratch.
 */

static void
remove_liberty(int str, int i, int j)
{
  int k;
  struct string_data *s = &string[str];
  
  if (s->liberties > MAX_LIBERTIES)
    update_liberties(str);
  else {
    for (k=0; k<s->liberties; k++)
      if (s->libi[k] == i && s->libj[k] == j) {
	/* We need to push the last entry too because it may become
	 * destroyed later.
	 */
	PUSH_VALUE(s->libi[s->liberties - 1]);
	PUSH_VALUE(s->libj[s->liberties - 1]);
	PUSH_VALUE(s->libi[k]);
	PUSH_VALUE(s->libj[k]);
	PUSH_VALUE(s->liberties);
	s->libi[k] = s->libi[s->liberties - 1];
	s->libj[k] = s->libj[s->liberties - 1];
	s->liberties--;
	break;
      }
  }
}


/* Remove a string from the board, pushing necessary information to
 * restore it. Return the number of removed stones.
 */

static int
incremental_remove_string(int s)
{
  int i, j, k;

  /* Traverse the stones of the string, by following the cyclic chain. */
  FIRST_STONE(s, i, j);
  do {
    /* Push color, string number and cyclic chain pointers. */
    PUSH_VERTEX(p[i][j]);
    PUSH_VALUE(string_number[i][j]);
    PUSH_VALUE(next_stonei[i][j]);
    PUSH_VALUE(next_stonej[i][j]);
    DO_REMOVE_STONE(i, j);
    NEXT_STONE(i, j);
  } while (!BACK_TO_FIRST_STONE(s, i, j));

  /* The neighboring strings have obtained some new liberties and lost
   * a neighbor.
   */
  for (k=0; k<string[s].neighbors; k++) {
    remove_neighbor(string[s].neighborlist[k], s);
    update_liberties(string[s].neighborlist[k]);
  }

  /* Update the number of captured stones. These are assumed to
   * already have been pushed.
   */
  if (string[s].color == WHITE)
    white_captured += string[s].size;
  else
    black_captured += string[s].size;
    
  return string[s].size;
}


/* Test whether a move would be a suicide.
 *
 * This is the case if
 *
 * 1. There is no neighboring empty intersection.
 * 2. There is no neighboring opponent string with exactly one liberty.
 * 3. There is no neighboring friendly string with more than one liberty.
 *
 */

int
incremental_is_suicide(int i, int j, int color)
{
  if (!strings_initialized)
    incremental_board_init();
  
  if (i > 0
      && (LIBERTY(i-1, j)
	  || ((p[i-1][j] == color) ^ (LIBERTIES(i-1, j) == 1))))
    return 0;

  if (i < board_size-1
      && (LIBERTY(i+1, j)
	  || ((p[i+1][j] == color) ^ (LIBERTIES(i+1, j) == 1))))
    return 0;

  if (j > 0
      && (LIBERTY(i, j-1)
	  || ((p[i][j-1] == color) ^ (LIBERTIES(i, j-1) == 1))))
    return 0;

  if (j < board_size-1
      && (LIBERTY(i, j+1)
	  || ((p[i][j+1] == color) ^ (LIBERTIES(i, j+1) == 1))))
    return 0;

  return 1;
}


/* We have played an isolated new stone and need to create a new
 * string for it.
 */
static void
create_new_string(int i, int j)
{
  int s;
  int color = p[i][j];

  /* Get the next free string number. */
  PUSH_VALUE(next_string);
  s = next_string++;
  string_number[i][j] = s;
  /* Set up a size one cycle for the string. */
  next_stonei[i][j] = i;
  next_stonej[i][j] = j;

  /* Set trivially known values and initialize the rest to zero. */
  string[s].color = color;
  string[s].size = 1;
  string[s].origini = i;
  string[s].originj = j;
  string[s].liberties = 0;
  string[s].neighbors = 0;
  string[s].mark = 0;

  /* Clear the string mark. */
  string_mark++;

  /* In each direction, look for a liberty or a nonmarked opponent
   * neighbor. Mark visited neighbors. There is no need to mark the
   * liberties since we can't find them twice. */
  if (i > 0) {
    if (LIBERTY(i-1, j)) {
      ADD_LIBERTY(s, i-1, j);
    }
    else if (UNMARKED_OPPONENT_STRING(s, i-1, j)) {
      int s2 = string_number[i-1][j];
      /* Add the neighbor to our list. */
      ADD_NEIGHBOR(s, i-1, j);
      /* Add us to our neighbor's list. */
      PUSH_VALUE(string[s2].neighbors);
      ADD_NEIGHBOR(s2, i, j);
      MARK_STRING(i-1, j);
    }
  }
  
  if (i < board_size-1) {
    if (LIBERTY(i+1, j)) {
      ADD_LIBERTY(s, i+1, j);
    }
    else if (UNMARKED_OPPONENT_STRING(s, i+1, j)) {
      int s2 = string_number[i+1][j];
      ADD_NEIGHBOR(s, i+1, j);
      PUSH_VALUE(string[s2].neighbors);
      ADD_NEIGHBOR(s2, i, j);
      MARK_STRING(i+1, j);
    }
  }
  
  if (j > 0) {
    if (LIBERTY(i, j-1)) {
      ADD_LIBERTY(s, i, j-1);
    }
    else if (UNMARKED_OPPONENT_STRING(s, i, j-1)) {
      int s2 = string_number[i][j-1];
      ADD_NEIGHBOR(s, i, j-1);
      PUSH_VALUE(string[s2].neighbors);
      ADD_NEIGHBOR(s2, i, j);
      MARK_STRING(i, j-1);
    }
  }
  
  if (j < board_size-1) {
    if (LIBERTY(i, j+1)) {
      ADD_LIBERTY(s, i, j+1);
    }
    else if (UNMARKED_OPPONENT_STRING(s, i, j+1)) {
      int s2 = string_number[i][j+1];
      ADD_NEIGHBOR(s, i, j+1);
      PUSH_VALUE(string[s2].neighbors);
      ADD_NEIGHBOR(s2, i, j);
      /* No need to mark since no visits left. */
/*      MARK_STRING(i, j+1);*/
    }
  }
}


/* We have played a stone with exactly one friendly neighbor. Add the
 * new stone to that string.
 */
static void
extend_neighbor_string(int i, int j, int s)
{
  int k;

  /* Link in the stone in the cyclic list. */
  int m = string[s].origini;
  int n = string[s].originj;
  next_stonei[i][j] = next_stonei[m][n];
  next_stonej[i][j] = next_stonej[m][n];
  PUSH_VALUE(next_stonei[m][n]);
  PUSH_VALUE(next_stonej[m][n]);
  next_stonei[m][n] = i;
  next_stonej[m][n] = j;
  
  /* Do we need to update the origin? */
  if (i < m || (i == m && j < n)) {
    PUSH_VALUE(string[s].origini);
    PUSH_VALUE(string[s].originj);
    string[s].origini = i;
    string[s].originj = j;
  }
  string_number[i][j] = s;

  /* The size of the string has increased by one. */
  PUSH_VALUE(string[s].size);
  string[s].size++;

  /* The place of the new stone is no longer a liberty. */
  remove_liberty(s, i, j);
  
  /* Mark old liberties and neighbors of the string. */
  liberty_mark++;
  string_mark++;
  for (k=0; k<string[s].liberties; k++)
    MARK_LIBERTY(string[s].libi[k], string[s].libj[k]);
  for (k=0; k<string[s].neighbors; k++)
    string[string[s].neighborlist[k]].mark = string_mark;

  /* Look at the neighbor locations of (i, j) for new liberties and/or
   * neighbor strings.
   */

  /* FIXME: conditional_add_neighbour should be possible to make
   *        unconditional.
   *
   * FIXME2: This doesn't work right if more than MAX_LIBERTY liberties.
   */
  if (i > 0) {
    if (UNMARKED_LIBERTY(i-1, j)) {
      ADD_LIBERTY(s, i-1, j);
    }
    else if (UNMARKED_OPPONENT_STRING(s, i-1, j)) {
      int s2 = string_number[i-1][j];
      PUSH_VALUE(string[s].neighbors);
      ADD_NEIGHBOR(s, i-1, j);
      conditional_add_neighbor(s2, s);
      MARK_STRING(i-1, j);
    }
  }
  
  if (i < board_size-1) {
    if (UNMARKED_LIBERTY(i+1, j)) {
      ADD_LIBERTY(s, i+1, j);
    }
    else if (UNMARKED_OPPONENT_STRING(s, i+1, j)) {
      int s2 = string_number[i+1][j];
      PUSH_VALUE(string[s].neighbors);
      ADD_NEIGHBOR(s, i+1, j);
      conditional_add_neighbor(s2, s);
      MARK_STRING(i+1, j);
    }
  }
  
  if (j > 0) {
    if (UNMARKED_LIBERTY(i, j-1)) {
      ADD_LIBERTY(s, i, j-1);
    }
    else if (UNMARKED_OPPONENT_STRING(s, i, j-1)) {
      int s2 = string_number[i][j-1];
      PUSH_VALUE(string[s].neighbors);
      ADD_NEIGHBOR(s, i, j-1);
      conditional_add_neighbor(s2, s);
      MARK_STRING(i, j-1);
    }
  }
  
  if (j < board_size-1) {
    if (UNMARKED_LIBERTY(i, j+1)) {
      ADD_LIBERTY(s, i, j+1);
    }
    else if (UNMARKED_OPPONENT_STRING(s, i, j+1)) {
      int s2 = string_number[i][j+1];
      PUSH_VALUE(string[s].neighbors);
      ADD_NEIGHBOR(s, i, j+1);
      conditional_add_neighbor(s2, s);
/*      MARK_STRING(i, j+1);*/
    }
  }
}


/* Incorporate the string at (i, j) with the string s.
 */

static void
assimilate_string(int s, int i, int j)
{
  int k;
  int lasti, lastj;
  int s2 = string_number[i][j];
  string[s].size += string[s2].size;

  /* Walk through the s2 stones and change string number. Also pick up
   * the last stone in the cycle for later use.
   */
  FIRST_STONE(s2, i, j);
  do {
    PUSH_VALUE(string_number[i][j]);
    string_number[i][j] = s;
    lasti = i;
    lastj = j;
    NEXT_STONE(i, j);
  } while(!BACK_TO_FIRST_STONE(s2, i, j));

  /* Link the two cycles together. */
  {
    int m = string[s].origini;
    int n = string[s].originj;
    PUSH_VALUE(next_stonei[lasti][lastj]);
    PUSH_VALUE(next_stonej[lasti][lastj]);
    PUSH_VALUE(next_stonei[m][n]);
    PUSH_VALUE(next_stonej[m][n]);
    next_stonei[lasti][lastj] = next_stonei[m][n];
    next_stonej[lasti][lastj] = next_stonej[m][n];
    next_stonei[m][n] = string[s2].origini;
    next_stonej[m][n] = string[s2].originj;
    
    /* Do we need to update the origin? */
    if (string[s2].origini < m || (string[s2].origini == m
				   && string[s2].originj < n)) {
      string[s].origini = string[s2].origini;
      string[s].originj = string[s2].originj;
    }
  }

  /* Pick up the liberties of s2 that we don't already have.
   * It is assumed that the liberties of s have been marked before
   * this function is called.
   *
   * FIXME: Problem if more than MAX_LIBERTY liberties.
   *
   */
  for (k=0; k<string[s2].liberties; k++) {
    int m = string[s2].libi[k];
    int n = string[s2].libj[k];
    if (UNMARKED_LIBERTY(m, n)) {
      ADD_AND_MARK_LIBERTY(s, m, n);
    }
  }

  /* Remove s2 as neighbor to the neighbors of s2 and instead add s if
   * they don't already have added it. Also add the neighbors of s2 as
   * neighbors of s, unless they already have been added. The already
   * known neighbors of s are assumed to have been marked before this
   * function is called.
   */
  for (k=0; k<string[s2].neighbors; k++) {
    int t = string[s2].neighborlist[k];
    remove_neighbor(t, s2);
    if (string[t].mark != string_mark) {
      PUSH_VALUE(string[t].neighbors);
      string[t].neighborlist[string[t].neighbors++] = s;
      string[s].neighborlist[string[s].neighbors++] = t;
      string[t].mark = string_mark;
    }
  }
}


/* Create a new string for the stone at (i, j) and assimilate all
 * friendly neighbor strings.
 */

static void
assimilate_neighbor_strings(int i, int j)
{
  int s;
  int color = p[i][j];

  /* Get the next free string number. */
  PUSH_VALUE(next_string);
  s = next_string++;
  string_number[i][j] = s;
  /* Set up a size one cycle for the string. */
  next_stonei[i][j] = i;
  next_stonej[i][j] = j;
  
  /* Set trivially known values and initialize the rest to zero. */
  string[s].color = color;
  string[s].size = 1;
  string[s].origini = i;
  string[s].originj = j;
  string[s].liberties = 0;
  string[s].neighbors = 0;

  /* Clear the marks. */
  liberty_mark++;
  string_mark++;

  /* Mark ourselves. */
  string[s].mark = string_mark;

  /* Look in each direction for
   *
   * 1. liberty: Add if not already visited.
   * 2. opponent string: Add it among our neighbors and us among its
   *    neighbors, unless already visited.
   * 3. friendly string: Assimilate.
   */
  if (i > 0) {
    if (UNMARKED_LIBERTY(i-1, j)) {
      ADD_AND_MARK_LIBERTY(s, i-1, j);
    }
    else if (UNMARKED_OPPONENT_STRING(s, i-1, j)) {
      ADD_NEIGHBOR(s, i-1, j);
      PUSH_VALUE(string[string_number[i-1][j]].neighbors);
      ADD_NEIGHBOR(string_number[i-1][j], i, j);
      MARK_STRING(i-1, j);
    }
    else if (UNMARKED_OWN_STRING(s, i-1, j)) {
      assimilate_string(s, i-1, j);
    }
  }

  if (i < board_size-1) {
    if (UNMARKED_LIBERTY(i+1, j)) {
      ADD_AND_MARK_LIBERTY(s, i+1, j);
    }
    else if (UNMARKED_OPPONENT_STRING(s, i+1, j)) {
      ADD_NEIGHBOR(s, i+1, j);
      PUSH_VALUE(string[string_number[i+1][j]].neighbors);
      ADD_NEIGHBOR(string_number[i+1][j], i, j);
      MARK_STRING(i+1, j);
    }
    else if (UNMARKED_OWN_STRING(s, i+1, j)) {
      assimilate_string(s, i+1, j);
    }
  }

  if (j > 0) {
    if (UNMARKED_LIBERTY(i, j-1)) {
      ADD_AND_MARK_LIBERTY(s, i, j-1);
    }
    else if (UNMARKED_OPPONENT_STRING(s, i, j-1)) {
      ADD_NEIGHBOR(s, i, j-1);
      PUSH_VALUE(string[string_number[i][j-1]].neighbors);
      ADD_NEIGHBOR(string_number[i][j-1], i, j);
      MARK_STRING(i, j-1);
    }
    else if (UNMARKED_OWN_STRING(s, i, j-1)) {
      assimilate_string(s, i, j-1);
    }
  }

  if (j < board_size-1) {
    if (UNMARKED_LIBERTY(i, j+1)) {
      ADD_AND_MARK_LIBERTY(s, i, j+1);
    }
    else if (UNMARKED_OPPONENT_STRING(s, i, j+1)) {
      ADD_NEIGHBOR(s, i, j+1);
      PUSH_VALUE(string[string_number[i][j+1]].neighbors);
      ADD_NEIGHBOR(string_number[i][j+1], i, j);
      MARK_STRING(i, j+1);
    }
    else if (UNMARKED_OWN_STRING(s, i, j+1)) {
      assimilate_string(s, i, j+1);
    }
  }

  /* If we now think we have exactly MAX_LIBERTIES-1 liberties, there
   * is a potential risk that we are missing a liberty, so we rebuild
   * the liberty list from scratch.
   *
   * An example of how this could occur is given below. Assume
   * MAX_LIBERTIES is 6 and we have the position
   *
   * ...XXXXX
   * XXXXdefX
   * XO*OOOOX
   * XXXXabcX
   * ...XXXXX
   *
   * The liberties are marked *,a,b,c,d,e,f and O plays at *. Before
   * the move the right string had 7 liberties, but only 6 could be
   * stored, assume those are *,a,b,c,d,e. After running the algorithm
   * up to this point, the newly created string at * will think it
   * only has the five liberties a,b,c,d,e.
   *
   * FIXME: We should probably change this strategy.
   */

  if (string[s].liberties == MAX_LIBERTIES-1) {
    string[s].liberties = 0; /* To avoiding pushing the current list. */
    update_liberties(s);
  }
}


/* Suicide at (i, j). Remove the neighboring friendly strings.
 */

static void
incremental_commit_suicide(int i, int j, int color)
{
  if (i > 0 && p[i-1][j] == color)
    incremental_remove_string(string_number[i-1][j]);

  if (i < board_size-1 && p[i+1][j] == color)
    incremental_remove_string(string_number[i+1][j]);

  if (j > 0 && p[i][j-1] == color)
    incremental_remove_string(string_number[i][j-1]);

  if (j < board_size-1 && p[i][j+1] == color)
    incremental_remove_string(string_number[i][j+1]);
}


/* Play a move without legality checking. Suicide is allowed. */
void
incremental_play_move(int i, int j, int color)
{
  int other = OTHER_COLOR(color);
  int captured_stones = 0;
  int neighbor_allies = 0;
  int have_liberties = 0;
  int s = -1;
  
  if (!strings_initialized)
    incremental_board_init();
    
  /* Remove captured stones and check for suicide.*/
  if (i > 0) {
    if (p[i-1][j] == other && LIBERTIES(i-1, j) == 1)
      captured_stones += incremental_remove_string(string_number[i-1][j]);
    else if (LIBERTY(i-1, j) || (p[i-1][j] == color && LIBERTIES(i-1, j) > 1))
      have_liberties = 1;
  }

  if (i < board_size-1) {
    if (p[i+1][j] == other && LIBERTIES(i+1, j) == 1)
      captured_stones += incremental_remove_string(string_number[i+1][j]);
    else if (LIBERTY(i+1, j) || (p[i+1][j] == color && LIBERTIES(i+1, j) > 1))
      have_liberties = 1;
  }

  if (j > 0) {
    if (p[i][j-1] == other && LIBERTIES(i, j-1) == 1)
      captured_stones += incremental_remove_string(string_number[i][j-1]);
    else if (LIBERTY(i, j-1) || (p[i][j-1] == color && LIBERTIES(i, j-1) > 1))
      have_liberties = 1;
  }

  if (j < board_size-1) {
    if (p[i][j+1] == other && LIBERTIES(i, j+1) == 1)
      captured_stones += incremental_remove_string(string_number[i][j+1]);
    else if (LIBERTY(i, j+1) || (p[i][j+1] == color && LIBERTIES(i, j+1) > 1))
      have_liberties = 1;
  }

  if (have_liberties == 0 && captured_stones == 0) {
    incremental_commit_suicide(i, j, color);
    return;
  }
  
  /* Put down the stone. */
  DO_ADD_STONE(i, j, color);

  /* Count the number of adjacent strings of my color and remove
   * (i, j) as liberty for the adjacent opponent strings.
   */
  string_mark++;
  if (i > 0 && p[i-1][j] != EMPTY && UNMARKED_STRING(i-1, j)) {
    if (p[i-1][j] == color) {
      neighbor_allies++;
      s = string_number[i-1][j];
    }
    else
      remove_liberty(string_number[i-1][j], i, j);
    MARK_STRING(i-1, j);
  }
  
  if (i < board_size-1 && p[i+1][j] != EMPTY && UNMARKED_STRING(i+1, j)) {
    if (p[i+1][j] == color) {
      neighbor_allies++;
      s = string_number[i+1][j];
    }
    else
      remove_liberty(string_number[i+1][j], i, j);
    MARK_STRING(i+1, j);
  }
  
  if (j > 0 && p[i][j-1] != EMPTY && UNMARKED_STRING(i, j-1)) {
    if (p[i][j-1] == color) {
      neighbor_allies++;
      s = string_number[i][j-1];
    }
    else
      remove_liberty(string_number[i][j-1], i, j);
    MARK_STRING(i, j-1);
  }
  
  if (j < board_size-1 && p[i][j+1] != EMPTY && UNMARKED_STRING(i, j+1)) {
    if (p[i][j+1] == color) {
      neighbor_allies++;
      s = string_number[i][j+1];
    }
    else
      remove_liberty(string_number[i][j+1], i, j);
/*    MARK_STRING(i, j+1);*/
  }
  
  if (neighbor_allies == 0)
    create_new_string(i, j);
  else if (neighbor_allies == 1) {
    assert(s >= 0);
    extend_neighbor_string(i, j, s);
  }
  else
    assimilate_neighbor_strings(i, j);

  s = string_number[i][j];
  
  /* Check whether this was a ko capture and if so set (ko_i, ko_j). */
  if (string[s].liberties == 1
      && string[s].size == 1
      && captured_stones == 1) {
    ko_i = string[s].libi[0];
    ko_j = string[s].libj[0];
    hashdata_set_ko(&hashdata, ko_i, ko_j);
  }
}

void
incremental_undo_move()
{
  assert(strings_initialized);
  if (0)
    gprintf("Change stack size = %d\n", change_stack_index);
  POP_MOVE();
  POP_VERTICES();
}

int
incremental_countlib(int i, int j, int color, int maxlib)
{
  int libs;
  
  if (!strings_initialized)
    incremental_board_init();
  
  if (p[i][j] != EMPTY)
    libs = string[string_number[i][j]].liberties;
  else {
    int l;
    libs = 0;
    liberty_mark++;
    MARK_LIBERTY(i, j);
    
    if (i > 0) {
      if (UNMARKED_LIBERTY(i-1, j)) {
	MARK_LIBERTY(i-1, j);
	libs++;if(libs>=maxlib)return libs;
      }
      else if (p[i-1][j] == color) {
	int s = string_number[i-1][j];
	for (l=0; l<string[s].liberties; l++) {
	  int ai = string[s].libi[l];
	  int aj = string[s].libj[l];
	  if (UNMARKED_LIBERTY(ai, aj)) {
	    MARK_LIBERTY(ai, aj);
	    libs++;if(libs>=maxlib)return libs;
	  }
	}
      }
    }
    
    if (i < board_size-1) {
      if (UNMARKED_LIBERTY(i+1, j)) {
	MARK_LIBERTY(i+1, j);
	libs++;if(libs>=maxlib)return libs;
      }
      else if (p[i+1][j] == color) {
	int s = string_number[i+1][j];
	for (l=0; l<string[s].liberties; l++) {
	  int ai = string[s].libi[l];
	  int aj = string[s].libj[l];
	  if (UNMARKED_LIBERTY(ai, aj)) {
	    MARK_LIBERTY(ai, aj);
	    libs++;if(libs>=maxlib)return libs;
	  }
	}
      }
    }
    
    if (j > 0) {
      if (UNMARKED_LIBERTY(i, j-1)) {
	MARK_LIBERTY(i, j-1);
	libs++;if(libs>=maxlib)return libs;
      }
      else if (p[i][j-1] == color) {
	int s = string_number[i][j-1];
	for (l=0; l<string[s].liberties; l++) {
	  int ai = string[s].libi[l];
	  int aj = string[s].libj[l];
	  if (UNMARKED_LIBERTY(ai, aj)) {
	    MARK_LIBERTY(ai, aj);
	    libs++;if(libs>=maxlib)return libs;
	  }
	}
      }
    }
    
    if (j < board_size-1) {
      if (UNMARKED_LIBERTY(i, j+1)) {
	MARK_LIBERTY(i, j+1);
	libs++;if(libs>=maxlib)return libs;
      }
      else if (p[i][j+1] == color) {
	int s = string_number[i][j+1];
	for (l=0; l<string[s].liberties; l++) {
	  int ai = string[s].libi[l];
	  int aj = string[s].libj[l];
	  if (UNMARKED_LIBERTY(ai, aj)) {
	    MARK_LIBERTY(ai, aj);
	    libs++;if(libs>=maxlib)return libs;
	  }
	}
      }
    }
  }
  return libs;
}

int
incremental_findlib(int i, int j, int color, int maxlib, int *libi, int *libj)
{
  int k;
  int libs;
  
  assert(libi != NULL && libj != NULL);

  /* FIXME: This is fishy. */
#if 0
  assert(maxlib <= MAX_LIBERTIES); /* The other case is not implemented yet. */
#else
  if (maxlib > MAX_LIBERTIES)
    maxlib = MAX_LIBERTIES;
#endif
  
  if (!strings_initialized)
    incremental_board_init();
  
  if (p[i][j] != EMPTY) {
    int s = string_number[i][j];
    libs = string[s].liberties;
    for (k=0; k<maxlib && k<libs; k++) {
      libi[k] = string[s].libi[k];
      libj[k] = string[s].libj[k];
    }
  }
  else {
    int l;
    libs = 0;
    liberty_mark++;
    MARK_LIBERTY(i, j);
    
    if (i > 0) {
      if (UNMARKED_LIBERTY(i-1, j)) {
	if (libs < maxlib) {
	  libi[libs] = i-1;
	  libj[libs] = j;
	}
	MARK_LIBERTY(i-1, j);
	libs++;
      }
      else if (p[i-1][j] == color) {
	int s = string_number[i-1][j];
	for (l=0; l<string[s].liberties; l++) {
	  int ai = string[s].libi[l];
	  int aj = string[s].libj[l];
	  if (UNMARKED_LIBERTY(ai, aj)) {
	    if (libs < maxlib) {
	      libi[libs] = ai;
	      libj[libs] = aj;
	    }
	    MARK_LIBERTY(ai, aj);
	    libs++;
	  }
	}
      }
    }
    
    if (i < board_size-1) {
      if (UNMARKED_LIBERTY(i+1, j)) {
	if (libs < maxlib) {
	  libi[libs] = i+1;
	  libj[libs] = j;
	}
	MARK_LIBERTY(i+1, j);
	libs++;
      }
      else if (p[i+1][j] == color) {
	int s = string_number[i+1][j];
	for (l=0; l<string[s].liberties; l++) {
	  int ai = string[s].libi[l];
	  int aj = string[s].libj[l];
	  if (UNMARKED_LIBERTY(ai, aj)) {
	    if (libs < maxlib) {
	      libi[libs] = ai;
	      libj[libs] = aj;
	    }
	    MARK_LIBERTY(ai, aj);
	    libs++;
	  }
	}
      }
    }
    
    if (j > 0) {
      if (UNMARKED_LIBERTY(i, j-1)) {
	if (libs < maxlib) {
	  libi[libs] = i;
	  libj[libs] = j-1;
	}
	MARK_LIBERTY(i, j-1);
	libs++;
      }
      else if (p[i][j-1] == color) {
	int s = string_number[i][j-1];
	for (l=0; l<string[s].liberties; l++) {
	  int ai = string[s].libi[l];
	  int aj = string[s].libj[l];
	  if (UNMARKED_LIBERTY(ai, aj)) {
	    if (libs < maxlib) {
	      libi[libs] = ai;
	      libj[libs] = aj;
	    }
	    MARK_LIBERTY(ai, aj);
	    libs++;
	  }
	}
      }
    }
    
    if (j < board_size-1) {
      if (UNMARKED_LIBERTY(i, j+1)) {
	if (libs < maxlib) {
	  libi[libs] = i;
	  libj[libs] = j+1;
	}
	MARK_LIBERTY(i, j+1);
	libs++;
      }
      else if (p[i][j+1] == color) {
	int s = string_number[i][j+1];
	for (l=0; l<string[s].liberties; l++) {
	  int ai = string[s].libi[l];
	  int aj = string[s].libj[l];
	  if (UNMARKED_LIBERTY(ai, aj)) {
	    if (libs < maxlib) {
	      libi[libs] = ai;
	      libj[libs] = aj;
	    }
	    MARK_LIBERTY(ai, aj);
	    libs++;
	  }
	}
      }
    }
  }

  return libs;
}

int
incremental_countsize(int i, int j)
{
  if (!strings_initialized)
    incremental_board_init();

  return string[string_number[i][j]].size;
}

void 
incremental_chainlinks(int i, int j, int *adj, int adji[MAXCHAIN],
		       int adjj[MAXCHAIN], int adjsize[MAXCHAIN],
		       int adjlib[MAXCHAIN])
{
  struct string_data *s, *t;
  int k;
  ASSERT(p[i][j] != EMPTY, i, j);

  if (!strings_initialized)
    incremental_board_init();
  
  s = &string[string_number[i][j]];
  for (k=0; k<s->neighbors; k++) {
    t = &string[s->neighborlist[k]];
    adji[k] = t->origini;
    adjj[k] = t->originj;
    adjsize[k] = t->size;
    adjlib[k] = t->liberties;
  }
  *adj = s->neighbors;
}

void
incremental_find_origin(int i, int j, int *origini, int *originj)
{
  struct string_data *s;

  if (!strings_initialized)
    incremental_board_init();
  
  s = &string[string_number[i][j]];
  *origini = s->origini;
  *originj = s->originj;
}

/*
 * Return true if the move (i,j) by (color) is a ko capture
 * (whether capture is legal on this move or not).
 */
int
incremental_is_ko(int i, int j, int color)
{
  int other = OTHER_COLOR(color);
  int captures = 0;
  
  if (!strings_initialized)
    incremental_board_init();
  
  if (i > 0) {
    if (p[i-1][j] != other)
      return 0;
    else if (LIBERTIES(i-1, j) == 1) {
      captures += string[string_number[i-1][j]].size;
      if (captures > 1)
	return 0;
    }
  }
  
  if (i < board_size-1) {
    if (p[i+1][j] != other)
      return 0;
    else if (LIBERTIES(i+1, j) == 1) {
      captures += string[string_number[i+1][j]].size;
      if (captures > 1)
	return 0;
    }
  }
  
  if (j > 0) {
    if (p[i][j-1] != other)
      return 0;
    else if (LIBERTIES(i, j-1) == 1) {
      captures += string[string_number[i][j-1]].size;
      if (captures > 1)
	return 0;
    }
  }
  
  if (j < board_size-1) {
    if (p[i][j+1] != other)
      return 0;
    else if (LIBERTIES(i, j+1) == 1) {
      captures += string[string_number[i][j+1]].size;
      if (captures > 1)
	return 0;
    }
  }

  return (captures == 1);
}

/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */
